@@ -14,26 +14,35 @@
 #    under the License.
 
 from eventlet import greenthread
+import netaddr
+from neutron.agent import firewall
+from os_win import exceptions
+from os_win.utils.network import networkutils
+from os_win import utilsfactory
 from oslo_log import log as logging
 import six
 
 from hyperv.common.i18n import _LE, _LI  # noqa
-from hyperv.neutron import utilsfactory
-from hyperv.neutron import utilsv2
+import threading
 
 LOG = logging.getLogger(__name__)
 
+INGRESS_DIRECTION = 'ingress'
+EGRESS_DIRECTION = 'egress'
+DIRECTION_IP_PREFIX = {'ingress': 'source_ip_prefix',
+                       'egress': 'dest_ip_prefix'}
+
 ACL_PROP_MAP = {
-    'direction': {'ingress': utilsv2.HyperVUtilsV2._ACL_DIR_IN,
-                  'egress': utilsv2.HyperVUtilsV2._ACL_DIR_OUT},
-    'ethertype': {'IPv4': utilsv2.HyperVUtilsV2._ACL_TYPE_IPV4,
-                  'IPv6': utilsv2.HyperVUtilsV2._ACL_TYPE_IPV6},
-    'protocol': {'tcp': utilsv2.HyperVUtilsV2._TCP_PROTOCOL,
-                 'udp': utilsv2.HyperVUtilsV2._UDP_PROTOCOL,
-                 'icmp': utilsv2.HyperVUtilsV2._ICMP_PROTOCOL,
-                 'icmpv6': utilsv2.HyperVUtilsV2._ICMPV6_PROTOCOL},
-    'action': {'allow': utilsv2.HyperVUtilsV2._ACL_ACTION_ALLOW,
-               'deny': utilsv2.HyperVUtilsV2._ACL_ACTION_DENY},
+    'direction': {'ingress': networkutils.NetworkUtils._ACL_DIR_IN,
+                  'egress': networkutils.NetworkUtils._ACL_DIR_OUT},
+    'ethertype': {'IPv4': networkutils.NetworkUtils._ACL_TYPE_IPV4,
+                  'IPv6': networkutils.NetworkUtils._ACL_TYPE_IPV6},
+    'protocol': {'tcp': networkutils.NetworkUtils._TCP_PROTOCOL,
+                 'udp': networkutils.NetworkUtils._UDP_PROTOCOL,
+                 'icmp': networkutils.NetworkUtils._ICMP_PROTOCOL,
+                 'ipv6-icmp': networkutils.NetworkUtils._ICMPV6_PROTOCOL},
+    'action': {'allow': networkutils.NetworkUtils._ACL_ACTION_ALLOW,
+               'deny': networkutils.NetworkUtils._ACL_ACTION_DENY},
     'default': "ANY",
     'address_default': {'IPv4': '0.0.0.0/0', 'IPv6': '::/0'}
 }
@@ -46,10 +55,73 @@ class HyperVSecurityGroupsDriverMixin(object):
     """
 
     def __init__(self):
-        self._utils = utilsfactory.get_hypervutils()
+        self._utils = utilsfactory.get_networkutils()
         self._sg_gen = SecurityGroupRuleGeneratorR2()
         self._sec_group_rules = {}
         self._security_ports = {}
+        self._sg_members = {}
+        self._sg_rule_templates = {}
+        self.cache_lock = threading.Lock()
+
+        # TODO(claudiub): remove this on the next os-win release.
+        clear_cache = lambda port_id: self._utils._sg_acl_sds.pop(port_id,
+                                                                  None)
+        self._utils.clear_port_sg_acls_cache = clear_cache
+
+    def _select_sg_rules_for_port(self, port, direction):
+        sg_ids = port.get('security_groups', [])
+        port_rules = []
+        fixed_ips = port.get('fixed_ips', [])
+        for sg_id in sg_ids:
+            for rule in self._sg_rule_templates.get(sg_id, []):
+                if rule['direction'] != direction:
+                    continue
+                remote_group_id = rule.get('remote_group_id')
+                if not remote_group_id:
+                    grp_rule = rule.copy()
+                    grp_rule['security_group_id'] = sg_id
+                    port_rules.append(grp_rule)
+                    continue
+                ethertype = rule['ethertype']
+                for ip in self._sg_members[remote_group_id][ethertype]:
+                    if ip in fixed_ips:
+                        continue
+                    ip_rule = rule.copy()
+                    direction_ip_prefix = DIRECTION_IP_PREFIX[direction]
+                    ip_rule[direction_ip_prefix] = str(
+                        netaddr.IPNetwork(ip).cidr)
+                    ip_rule['security_group_id'] = sg_id
+                    port_rules.append(ip_rule)
+        return port_rules
+
+    def filter_defer_apply_on(self):
+        """Defer application of filtering rule."""
+        pass
+
+    def filter_defer_apply_off(self):
+        """Turn off deferral of rules and apply the rules now."""
+        pass
+
+    def update_security_group_rules(self, sg_id, sg_rules):
+        LOG.debug("Update rules of security group (%s)", sg_id)
+        with self.cache_lock:
+            self._sg_rule_templates[sg_id] = sg_rules
+
+    def update_security_group_members(self, sg_id, sg_members):
+        LOG.debug("Update members of security group (%s)", sg_id)
+        with self.cache_lock:
+            self._sg_members[sg_id] = sg_members
+
+    def _generate_rules(self, ports):
+        newports = {}
+        for port in ports:
+            _rules = []
+            _rules.extend(self._select_sg_rules_for_port(port,
+                                                         INGRESS_DIRECTION))
+            _rules.extend(self._select_sg_rules_for_port(port,
+                                                         EGRESS_DIRECTION))
+            newports[port['id']] = _rules
+        return newports
 
     def prepare_port_filter(self, port):
         LOG.debug('Creating port %s rules', len(port['security_group_rules']))
@@ -61,9 +133,15 @@ class HyperVSecurityGroupsDriverMixin(object):
 
             def_sg_rules = self._sg_gen.create_default_sg_rules()
             self._add_sg_port_rules(port['id'], def_sg_rules)
+            # Add provider rules
+            provider_rules = port['security_group_rules']
+            self._create_port_rules(port['id'], provider_rules)
+
+        newrules = self._generate_rules([port])
+        self._create_port_rules(port['id'], newrules[port['id']])
 
         self._security_ports[port['device']] = port
-        self._create_port_rules(port['id'], port['security_group_rules'])
+        self._sec_group_rules[port['id']] = newrules[port['id']]
 
     def _create_port_rules(self, port_id, rules):
         sg_rules = self._sg_gen.create_security_group_rules(rules)
@@ -86,10 +164,14 @@ class HyperVSecurityGroupsDriverMixin(object):
         try:
             self._utils.create_security_rules(port_id, sg_rules)
             old_sg_rules.extend(sg_rules)
-        except Exception as ex:
-            LOG.error(_LE('Hyper-V Exception: %(hyperv_exeption)s while '
-                          'adding rules for port: %(port_id)s'),
-                      dict(hyperv_exeption=ex, port_id=port_id))
+        except exceptions.NotFound:
+            # port no longer exists.
+            self._sec_group_rules.pop(port_id, None)
+            raise
+        except Exception:
+            LOG.exception(_LE('Exception encountered while adding rules for '
+                              'port: %s'), port_id)
+            raise
 
     def _remove_sg_port_rules(self, port_id, sg_rules):
         if not sg_rules:
@@ -98,11 +180,16 @@ class HyperVSecurityGroupsDriverMixin(object):
         try:
             self._utils.remove_security_rules(port_id, sg_rules)
             for rule in sg_rules:
-                old_sg_rules.remove(rule)
-        except Exception as ex:
-            LOG.error(_LE('Hyper-V Exception: %(hyperv_exeption)s while '
-                          'removing rules for port: %(port_id)s'),
-                      dict(hyperv_exeption=ex, port_id=port_id))
+                if rule in old_sg_rules:
+                    old_sg_rules.remove(rule)
+        except exceptions.NotFound:
+            # port no longer exists.
+            self._sec_group_rules.pop(port_id, None)
+            raise
+        except Exception:
+            LOG.exception(_LE('Exception encountered while removing rules for '
+                              'port: %s'), port_id)
+            raise
 
     def apply_port_filter(self, port):
         LOG.info(_LI('Aplying port filter.'))
@@ -111,29 +198,43 @@ class HyperVSecurityGroupsDriverMixin(object):
         LOG.info(_LI('Updating port rules.'))
 
         if port['device'] not in self._security_ports:
-            self.prepare_port_filter(port)
+            LOG.info(_LI("Device %(port)s not yet added."),
+                     {'port': port['id']})
             return
 
         old_port = self._security_ports[port['device']]
-        rules = old_port['security_group_rules']
-        param_port_rules = port['security_group_rules']
-
-        new_rules = [r for r in param_port_rules if r not in rules]
-        remove_rules = [r for r in rules if r not in param_port_rules]
-
-        LOG.info(_LI("Creating %(new)s new rules, removing %(old)s "
-                     "old rules."),
-                 {'new': len(new_rules), 'old': len(remove_rules)})
+        old_provider_rules = old_port['security_group_rules']
+        added_provider_rules = port['security_group_rules']
+        # Generate the rules
+        added_rules = self._generate_rules([port])
+        # Consider added provider rules (if any)
+        new_rules = [r for r in added_provider_rules
+                     if r not in old_provider_rules]
+        # Build new rules to add
+        new_rules.extend([r for r in added_rules[port['id']]
+                          if r not in self._sec_group_rules[port['id']]])
+        # Remove non provider rules
+        remove_rules = [r for r in self._sec_group_rules[port['id']]
+                        if r not in added_rules[port['id']]]
+        # Remove for non provider rules
+        remove_rules.extend([r for r in old_provider_rules
+                             if r not in added_provider_rules])
+        LOG.info(_("Creating %(new)s new rules, removing %(old)s "
+                   "old rules."),
+                 {'new': len(new_rules),
+                  'old': len(remove_rules)})
 
         self._create_port_rules(port['id'], new_rules)
         self._remove_port_rules(old_port['id'], remove_rules)
 
         self._security_ports[port['device']] = port
+        self._sec_group_rules[port['id']] = added_rules[port['id']]
 
     def remove_port_filter(self, port):
         LOG.info(_LI('Removing port filter'))
         self._security_ports.pop(port['device'], None)
         self._sec_group_rules.pop(port['id'], None)
+        self._utils.clear_port_sg_acls_cache(port['id'])
 
     def security_group_updated(self, action_type, sec_group_ids,
                                device_id=None):
@@ -261,7 +362,7 @@ class SecurityGroupRuleR2(SecurityGroupRuleBase):
     def __init__(self, direction, local_port, protocol, remote_addr,
                  action=ACL_PROP_MAP['action']['allow']):
         is_not_icmp = protocol not in [ACL_PROP_MAP['protocol']['icmp'],
-                                       ACL_PROP_MAP['protocol']['icmpv6']]
+                                       ACL_PROP_MAP['protocol']['ipv6-icmp']]
 
         self.Direction = direction
         self.Action = action
@@ -279,3 +380,8 @@ class SecurityGroupRuleR2(SecurityGroupRuleBase):
 
     def __hash__(self):
         return self._cached_hash
+
+
+class HyperVSecurityGroupsDriver(HyperVSecurityGroupsDriverMixin,
+                                 firewall.FirewallDriver):
+    pass
