@@ -18,13 +18,16 @@
 Unit tests for Windows Hyper-V virtual switch neutron driver
 """
 
+from concurrent import futures
+import time
+
 import mock
+from os_win import exceptions
+from os_win import utilsfactory
 
 from hyperv.neutron import constants
-from hyperv.neutron import hyperv_agent_notifier
+from hyperv.neutron import exception
 from hyperv.neutron import hyperv_neutron_agent
-from hyperv.neutron import utils
-from hyperv.neutron import utilsfactory
 from hyperv.tests import base
 
 
@@ -34,12 +37,13 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
 
     def setUp(self):
         super(TestHyperVNeutronAgent, self).setUp()
-
-        utilsfactory._get_windows_version = mock.MagicMock(
-            return_value='6.1.0')
+        utilsfactory_patcher = mock.patch.object(utilsfactory, '_get_class')
+        utilsfactory_patcher.start()
+        self.addCleanup(utilsfactory_patcher.stop)
 
         self.agent = hyperv_neutron_agent.HyperVNeutronAgentMixin()
         self.agent.plugin_rpc = mock.Mock()
+        self.agent._metricsutils = mock.MagicMock()
         self.agent._utils = mock.MagicMock()
         self.agent.sec_groups_agent = mock.MagicMock()
         self.agent.context = mock.Mock()
@@ -49,6 +53,23 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
         self.agent.notifier = mock.Mock()
         self.agent._utils = mock.MagicMock()
         self.agent._nvgre_ops = mock.MagicMock()
+        self.agent._workers = mock.MagicMock()
+
+    @mock.patch.object(hyperv_neutron_agent, 'synchronized')
+    def test_port_synchronized(self, mock_synchronized):
+        fake_method_side_effect = mock.Mock()
+
+        @hyperv_neutron_agent._port_synchronized
+        def fake_method(fake_arg, port_id):
+            fake_method_side_effect(fake_arg, port_id)
+
+        mock_synchronized.return_value = lambda x: x
+        expected_lock_name = 'port-lock-%s' % mock.sentinel.port_id
+
+        fake_method(fake_arg=mock.sentinel.arg, port_id=mock.sentinel.port_id)
+        mock_synchronized.assert_called_once_with(expected_lock_name)
+        fake_method_side_effect.assert_called_once_with(
+            mock.sentinel.arg, mock.sentinel.port_id)
 
     def test_load_physical_network_mappings(self):
         test_mappings = ['fakenetwork1:fake_vswitch',
@@ -71,7 +92,8 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
     @mock.patch.object(hyperv_neutron_agent.nvgre_ops, 'HyperVNvgreOps')
     def test_init_nvgre_no_tunnel_ip(self, mock_hyperv_nvgre_ops):
         self.config(enable_support=True, group='NVGRE')
-        self.assertRaises(utils.HyperVException, self.agent._init_nvgre)
+        self.assertRaises(exception.NetworkingHyperVException,
+                          self.agent._init_nvgre)
 
     @mock.patch.object(hyperv_neutron_agent.nvgre_ops, 'HyperVNvgreOps')
     def test_init_nvgre_enabled(self, mock_hyperv_nvgre_ops):
@@ -85,30 +107,6 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
         self.assertTrue(self.agent._nvgre_enabled)
         self.agent._nvgre_ops.init_notifier.assert_called_once_with(
             self.agent.context, self.agent.client)
-        expected_topic = hyperv_agent_notifier.get_topic_name(
-            self.agent.topic, constants.LOOKUP, constants.UPDATE)
-        self.agent.connection.create_consumer.assert_called_once_with(
-            expected_topic, self.agent.endpoints, fanout=True)
-        self.agent.connection.servers[-1].start.assert_called_once_with()
-
-    def test_get_agent_configurations(self):
-        actual = self.agent.get_agent_configurations()
-
-        self.assertEqual(self.agent._physical_network_mappings,
-                         actual['vswitch_mappings'])
-        self.assertNotIn('tunnel_types', actual)
-        self.assertNotIn('tunneling_ip', actual)
-
-    def test_get_agent_configurations_nvgre(self):
-        self.config(enable_support=True, group='NVGRE')
-        self.config(provider_tunnel_ip=mock.sentinel.tunneling_ip,
-                    group='NVGRE')
-        actual = self.agent.get_agent_configurations()
-
-        self.assertEqual(self.agent._physical_network_mappings,
-                         actual['vswitch_mappings'])
-        self.assertEqual([constants.TYPE_NVGRE], actual['tunnel_types'])
-        self.assertEqual(mock.sentinel.tunneling_ip, actual['tunneling_ip'])
 
     def test_get_network_vswitch_map_by_port_id(self):
         net_uuid = 'net-uuid'
@@ -132,6 +130,33 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
         self.assertIsNone(network)
         self.assertIsNone(port_map)
 
+    @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
+                       '_treat_vif_port')
+    def test_port_update_not_found(self, mock_treat_vif_port):
+        self.agent._utils.vnic_port_exists.return_value = False
+        port = {'id': mock.sentinel.port_id}
+        self.agent.port_update(self.agent.context, port)
+
+        self.assertFalse(mock_treat_vif_port.called)
+
+    @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
+                       '_treat_vif_port')
+    def test_port_update(self, mock_treat_vif_port):
+        self.agent._utils.vnic_port_exists.return_value = True
+        port = {'id': mock.sentinel.port_id,
+                'network_id': mock.sentinel.network_id,
+                'admin_state_up': mock.sentinel.admin_state_up}
+
+        self.agent.port_update(self.agent.context, port,
+                               mock.sentinel.network_type,
+                               mock.sentinel.segmentation_id,
+                               mock.sentinel.physical_network)
+
+        mock_treat_vif_port.assert_called_once_with(
+            mock.sentinel.port_id, mock.sentinel.network_id,
+            mock.sentinel.network_type, mock.sentinel.physical_network,
+            mock.sentinel.segmentation_id, mock.sentinel.admin_state_up)
+
     def test_lookup_update(self):
         kwargs = {'lookup_ip': mock.sentinel.lookup_ip,
                   'lookup_details': mock.sentinel.lookup_details}
@@ -160,7 +185,8 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
     @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
                        "_get_vswitch_name")
     def test_provision_network_exception(self, mock_get_vswitch_name):
-        self.assertRaises(utils.HyperVException, self.agent._provision_network,
+        self.assertRaises(exception.NetworkingHyperVException,
+                          self.agent._provision_network,
                           mock.sentinel.FAKE_PORT_ID,
                           mock.sentinel.FAKE_NET_UUID,
                           mock.sentinel.FAKE_NETWORK_TYPE,
@@ -173,7 +199,6 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
     @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
                        "_get_vswitch_name")
     def test_provision_network_vlan(self, mock_get_vswitch_name):
-        vswitch_name = mock_get_vswitch_name.return_value
         self.agent._provision_network(mock.sentinel.FAKE_PORT_ID,
                                       mock.sentinel.FAKE_NET_UUID,
                                       constants.TYPE_VLAN,
@@ -182,10 +207,6 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
         mock_get_vswitch_name.assert_called_once_with(
             constants.TYPE_VLAN,
             mock.sentinel.FAKE_PHYSICAL_NETWORK)
-        set_switch = self.agent._utils.set_switch_external_port_trunk_vlan
-        set_switch.assert_called_once_with(vswitch_name,
-                                           mock.sentinel.FAKE_SEGMENTATION_ID,
-                                           constants.TRUNK_ENDPOINT_MODE)
 
     @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
                        "_get_vswitch_name")
@@ -240,18 +261,12 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
 
     def _test_port_bound(self, enable_metrics):
         port = mock.MagicMock()
-        mock_enable_metrics = mock.MagicMock()
         net_uuid = 'my-net-uuid'
 
-        with mock.patch.multiple(
-                self.agent._utils,
-                connect_vnic_to_vswitch=mock.MagicMock(),
-                set_vswitch_port_vlan_id=mock.MagicMock(),
-                enable_port_metrics_collection=mock_enable_metrics):
+        self.agent._port_bound(port, net_uuid, 'vlan', None, None)
 
-            self.agent._port_bound(port, net_uuid, 'vlan', None, None)
-
-            self.assertEqual(enable_metrics, mock_enable_metrics.called)
+        self.assertEqual(enable_metrics,
+                         self.agent._utils.add_metrics_collection_acls.called)
 
     @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
                        '_provision_network')
@@ -295,14 +310,14 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
 
         with mock.patch.object(
                 self.agent._utils,
-                'disconnect_switch_port') as mock_disconnect_switch_port:
+                'remove_switch_port') as mock_remove_switch_port:
             self.agent._port_unbound(self._FAKE_PORT_ID, vnic_deleted=False)
 
             if net_uuid:
-                mock_disconnect_switch_port.assert_called_once_with(
-                    self._FAKE_PORT_ID, False, True)
+                mock_remove_switch_port.assert_called_once_with(
+                    self._FAKE_PORT_ID, False)
             else:
-                self.assertFalse(mock_disconnect_switch_port.called)
+                self.assertFalse(mock_remove_switch_port.called)
 
     @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
                        '_reclaim_local_network')
@@ -320,15 +335,12 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
         self.agent._port_metric_retries[self._FAKE_PORT_ID] = (
             self.agent._metrics_max_retries)
 
-        with mock.patch.multiple(self.agent._utils,
-                                 can_enable_control_metrics=mock.MagicMock(),
-                                 enable_control_metrics=mock.MagicMock()):
-
-            self.agent._utils.can_enable_control_metrics.return_value = True
-            self.agent._port_enable_control_metrics()
-            self.agent._utils.enable_control_metrics.assert_called_with(
-                self._FAKE_PORT_ID)
+        self.agent._utils.is_metrics_collection_allowed.return_value = True
+        self.agent._port_enable_control_metrics()
 
+        enable_port_metrics_collection = (
+            self.agent._metricsutils.enable_port_metrics_collection)
+        enable_port_metrics_collection.assert_called_with(self._FAKE_PORT_ID)
         self.assertNotIn(self._FAKE_PORT_ID, self.agent._port_metric_retries)
 
     def test_port_enable_control_metrics_maxed(self):
@@ -336,69 +348,192 @@ class TestHyperVNeutronAgent(base.BaseTestCase):
         self.agent._metrics_max_retries = 3
         self.agent._port_metric_retries[self._FAKE_PORT_ID] = 3
 
-        with mock.patch.multiple(self.agent._utils,
-                                 can_enable_control_metrics=mock.MagicMock(),
-                                 enable_control_metrics=mock.MagicMock()):
-
-            self.agent._utils.can_enable_control_metrics.return_value = False
-            for i in range(4):
-                self.assertIn(self._FAKE_PORT_ID,
-                              self.agent._port_metric_retries)
-                self.agent._port_enable_control_metrics()
+        self.agent._utils.is_metrics_collection_allowed.return_value = False
+        for i in range(4):
+            self.assertIn(self._FAKE_PORT_ID,
+                          self.agent._port_metric_retries)
+            self.agent._port_enable_control_metrics()
 
         self.assertNotIn(self._FAKE_PORT_ID, self.agent._port_metric_retries)
 
+    def test_port_enable_control_metrics_no_vnic(self):
+        self.agent.enable_metrics_collection = True
+        self.agent._port_metric_retries[self._FAKE_PORT_ID] = 3
+        self.agent._utils.is_metrics_collection_allowed.side_effect = (
+            exceptions.NotFound(resource=self._FAKE_PORT_ID))
+
+        self.agent._port_enable_control_metrics()
+        self.assertNotIn(self._FAKE_PORT_ID, self.agent._port_metric_retries)
+
+    @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
+                       '_port_unbound')
+    def test_vif_port_state_down(self, mock_port_unbound):
+        self.agent._treat_vif_port(
+            mock.sentinel.port_id, mock.sentinel.network_id,
+            mock.sentinel.network_type, mock.sentinel.physical_network,
+            mock.sentinel.segmentation_id, False)
+
+        mock_port_unbound.assert_called_once_with(mock.sentinel.port_id)
+        sg_agent = self.agent.sec_groups_agent
+        sg_agent.remove_devices_filter.assert_called_once_with(
+            [mock.sentinel.port_id])
+
+    @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
+                       '_port_bound')
+    def _check_treat_vif_port_state_up(self, mock_port_bound):
+        self.agent._treat_vif_port(
+            mock.sentinel.port_id, mock.sentinel.network_id,
+            mock.sentinel.network_type, mock.sentinel.physical_network,
+            mock.sentinel.segmentation_id, True)
+
+        mock_port_bound.assert_called_once_with(
+            mock.sentinel.port_id, mock.sentinel.network_id,
+            mock.sentinel.network_type, mock.sentinel.physical_network,
+            mock.sentinel.segmentation_id)
+
+    def test_treat_vif_port_sg_enabled(self):
+        self.agent.enable_security_groups = True
+        self._check_treat_vif_port_state_up()
+
+        sg_agent = self.agent.sec_groups_agent
+        sg_agent.prepare_devices_filter.assert_called_once_with(
+            [mock.sentinel.port_id])
+
+    def test_treat_vif_port_sg_disabled(self):
+        self.agent.enable_security_groups = False
+        self._check_treat_vif_port_state_up()
+        self.agent._utils.remove_all_security_rules.assert_called_once_with(
+            mock.sentinel.port_id)
+
+    @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
+                       '_treat_vif_port')
+    def test_process_added_port(self, mock_treat_vif_port):
+        self.agent._added_ports = set()
+        details = self._get_fake_port_details()
+
+        self.agent._process_added_port(details)
+
+        mock_treat_vif_port.assert_called_once_with(
+            mock.sentinel.port_id, mock.sentinel.network_id,
+            mock.sentinel.network_type, mock.sentinel.physical_network,
+            mock.sentinel.segmentation_id, mock.sentinel.admin_state_up)
+        self.agent.plugin_rpc.update_device_up.assert_called_once_with(
+            self.agent.context, mock.sentinel.device,
+            self.agent.agent_id, self.agent._host)
+        self.assertNotIn(mock.sentinel.device, self.agent._added_ports)
+
+    @mock.patch.object(hyperv_neutron_agent.HyperVNeutronAgentMixin,
+                       '_treat_vif_port')
+    def test_process_added_port_failed(self, mock_treat_vif_port):
+        mock_treat_vif_port.side_effect = exception.NetworkingHyperVException
+        self.agent._added_ports = set()
+        details = self._get_fake_port_details()
+
+        self.agent._process_added_port(details)
+        self.assertIn(mock.sentinel.device, self.agent._added_ports)
+
+    def _get_fake_port_details(self):
+        return {'device': mock.sentinel.device,
+                'port_id': mock.sentinel.port_id,
+                'network_id': mock.sentinel.network_id,
+                'network_type': mock.sentinel.network_type,
+                'physical_network': mock.sentinel.physical_network,
+                'segmentation_id': mock.sentinel.segmentation_id,
+                'admin_state_up': mock.sentinel.admin_state_up}
+
     def test_treat_devices_added_returns_true_for_missing_device(self):
+        self.agent._added_ports = set([mock.sentinel.port_id])
         attrs = {'get_devices_details_list.side_effect': Exception()}
         self.agent.plugin_rpc.configure_mock(**attrs)
-        self.assertTrue(self.agent._treat_devices_added([{}]))
+        self.agent._treat_devices_added()
 
-    def mock_treat_devices_added(self, details, func_name):
-        """Mock treat devices added.
-
-        :param details: the details to return for the device
-        :param func_name: the function that should be called
-        :returns: whether the named function was called
-        """
-        attrs = {'get_devices_details_list.return_value': [details]}
-        self.agent.plugin_rpc.configure_mock(**attrs)
-        with mock.patch.object(self.agent, func_name) as func:
-            self.assertFalse(self.agent._treat_devices_added([{}]))
-        return func.called
+        self.assertIn(mock.sentinel.port_id, self.agent._added_ports)
 
     def test_treat_devices_added_updates_known_port(self):
-        details = mock.MagicMock()
-        details.__contains__.side_effect = lambda x: True
-        with mock.patch.object(self.agent.plugin_rpc,
-                               "update_device_up") as func:
-            self.assertTrue(self.mock_treat_devices_added(details,
-                                                          '_treat_vif_port'))
-            self.assertTrue(func.called)
+        self.agent._added_ports = set([mock.sentinel.device])
+        details = self._get_fake_port_details()
+        attrs = {'get_devices_details_list.return_value': [details]}
+        self.agent.plugin_rpc.configure_mock(**attrs)
+
+        self.agent._treat_devices_added()
+
+        self.agent._workers.submit.assert_called_once_with(
+            self.agent._process_added_port, details)
+        self.assertNotIn(mock.sentinel.device, self.agent._added_ports)
 
     def test_treat_devices_added_missing_port_id(self):
-        details = mock.MagicMock()
-        details.__contains__.side_effect = lambda x: False
-        with mock.patch.object(self.agent.plugin_rpc,
-                               "update_device_up") as func:
-            self.assertFalse(self.mock_treat_devices_added(details,
-                                                           '_treat_vif_port'))
-            self.assertFalse(func.called)
+        self.agent._added_ports = set([mock.sentinel.port_id])
+        details = {'device': mock.sentinel.port_id}
+        attrs = {'get_devices_details_list.return_value': [details]}
+        self.agent.plugin_rpc.configure_mock(**attrs)
 
-    def test_treat_devices_removed_returns_true_for_missing_device(self):
+        self.agent._treat_devices_added()
+
+        self.assertFalse(self.agent._workers.submit.called)
+        self.assertNotIn(mock.sentinel.port_id, self.agent._added_ports)
+
+    def test_treat_devices_removed_exception(self):
+        self.agent._removed_ports = set([mock.sentinel.port_id])
         attrs = {'update_device_down.side_effect': Exception()}
         self.agent.plugin_rpc.configure_mock(**attrs)
-        self.assertTrue(self.agent._treat_devices_removed([{}]))
+        self.agent._treat_devices_removed()
+
+        self.agent.plugin_rpc.update_device_down.assert_called_once_with(
+            self.agent.context, mock.sentinel.port_id,
+            self.agent.agent_id, self.agent._host)
+        self.assertIn(mock.sentinel.port_id, self.agent._removed_ports)
 
     def mock_treat_devices_removed(self, port_exists):
+        self.agent._removed_ports = set([mock.sentinel.port_id])
         details = dict(exists=port_exists)
         attrs = {'update_device_down.return_value': details}
         self.agent.plugin_rpc.configure_mock(**attrs)
         with mock.patch.object(self.agent, '_port_unbound') as func:
-            self.assertFalse(self.agent._treat_devices_removed([{}]))
+            self.agent._treat_devices_removed()
         self.assertEqual(func.called, not port_exists)
+        self.assertEqual(
+            self.agent.sec_groups_agent.remove_devices_filter.called,
+            not port_exists)
+        self.assertNotIn(mock.sentinel.port_id, self.agent._removed_ports)
 
     def test_treat_devices_removed_unbinds_port(self):
         self.mock_treat_devices_removed(False)
 
     def test_treat_devices_removed_ignores_missing_port(self):
         self.mock_treat_devices_removed(False)
+
+    def test_process_added_port_event(self):
+        self.agent._added_ports = set()
+        self.agent._process_added_port_event(mock.sentinel.port_id)
+        self.assertIn(mock.sentinel.port_id, self.agent._added_ports)
+
+    def test_process_removed_port_event(self):
+        self.agent._removed_ports = set([])
+        self.agent._process_removed_port_event(mock.sentinel.port_id)
+        self.assertIn(mock.sentinel.port_id, self.agent._removed_ports)
+
+    @mock.patch.object(hyperv_neutron_agent.threading, 'Thread')
+    def test_create_event_listeners(self, mock_Thread):
+        self.agent._create_event_listeners()
+
+        self.agent._utils.get_vnic_event_listener.assert_has_calls([
+            mock.call(self.agent._utils.EVENT_TYPE_CREATE),
+            mock.call(self.agent._utils.EVENT_TYPE_DELETE)])
+        target = self.agent._utils.get_vnic_event_listener.return_value
+        calls = [mock.call(target=target,
+                           args=(self.agent._process_added_port_event, )),
+                 mock.call(target=target,
+                           args=(self.agent._process_removed_port_event, ))]
+        mock_Thread.assert_has_calls(calls, any_order=True)
+        self.assertEqual(2, mock_Thread.return_value.start.call_count)
+
+    def test_thread_pool_execution(self):
+        pool = futures.ThreadPoolExecutor(max_workers=3)
+        mock_fn = mock.MagicMock()
+
+        for i in range(8):
+            pool.submit(mock_fn, mock.sentinel.parameter)
+
+        # allow the threads to finish. one second is enough for a noop call.
+        time.sleep(1)
+        mock_fn.assert_has_calls([mock.call(mock.sentinel.parameter)] * 8)
