@@ -18,10 +18,11 @@ Unit tests for the Hyper-V Security Groups Driver.
 """
 
 import mock
+from os_win import exceptions
+from os_win import utilsfactory
 from oslo_config import cfg
 
 from hyperv.neutron import security_groups_driver as sg_driver
-from hyperv.neutron import utilsfactory
 from hyperv.tests import base
 
 CONF = cfg.CONF
@@ -35,7 +36,9 @@ class SecurityGroupRuleTestHelper(base.BaseTestCase):
     _FAKE_ACTION = sg_driver.ACL_PROP_MAP['action']['allow']
     _FAKE_DEST_IP_PREFIX = '10.0.0.0/24'
     _FAKE_SOURCE_IP_PREFIX = '10.0.1.0/24'
+    _FAKE_MEMBER_IP = '10.0.0.1'
     _FAKE_IPV6_LEN128_IP = 'fddd:cafd:e664:0:f816:3eff:fe8d:59d2/128'
+    _FAKE_SG_ID = 'fake_sg_id'
 
     _FAKE_PORT_MIN = 9001
     _FAKE_PORT_MAX = 9011
@@ -48,7 +51,8 @@ class SecurityGroupRuleTestHelper(base.BaseTestCase):
             'dest_ip_prefix': self._FAKE_DEST_IP_PREFIX,
             'source_ip_prefix': self._FAKE_SOURCE_IP_PREFIX,
             'port_range_min': self._FAKE_PORT_MIN,
-            'port_range_max': self._FAKE_PORT_MAX
+            'port_range_max': self._FAKE_PORT_MAX,
+            'security_group_id': self._FAKE_SG_ID
         }
 
     @classmethod
@@ -56,7 +60,7 @@ class SecurityGroupRuleTestHelper(base.BaseTestCase):
         return sg_driver.ACL_PROP_MAP[key1][key2]
 
 
-class TestHyperVSecurityGroupsDriverMixin(SecurityGroupRuleTestHelper):
+class TestHyperVSecurityGroupsDriver(SecurityGroupRuleTestHelper):
 
     _FAKE_DEVICE = 'fake_device'
     _FAKE_ID = 'fake_id'
@@ -64,39 +68,131 @@ class TestHyperVSecurityGroupsDriverMixin(SecurityGroupRuleTestHelper):
     _FAKE_PARAM_VALUE = 'fake_param_value'
 
     def setUp(self):
-        super(TestHyperVSecurityGroupsDriverMixin, self).setUp()
-        self._mock_windows_version = mock.patch.object(utilsfactory,
-                                                       'get_hypervutils')
-        self._mock_windows_version.start()
-        self._driver = sg_driver.HyperVSecurityGroupsDriverMixin()
+        super(TestHyperVSecurityGroupsDriver, self).setUp()
+        utilsfactory_patcher = mock.patch.object(utilsfactory, '_get_class')
+        utilsfactory_patcher.start()
+        self.addCleanup(utilsfactory_patcher.stop)
+
+        self._driver = sg_driver.HyperVSecurityGroupsDriver()
         self._driver._utils = mock.MagicMock()
         self._driver._sg_gen = mock.MagicMock()
 
-    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriverMixin,
+    def test__select_sg_rules_for_port(self):
+        mock_port = self._get_port()
+        mock_port['fixed_ips'] = [mock.MagicMock()]
+        mock_port['security_groups'] = [self._FAKE_SG_ID]
+
+        fake_sg_template = self._create_security_rule()
+        fake_sg_template['direction'] = 'ingress'
+        self._driver._sg_rule_templates[self._FAKE_SG_ID] = [fake_sg_template]
+
+        # Test without remote_group_id
+        rule_list = self._driver._select_sg_rules_for_port(mock_port,
+                                                           'ingress')
+        self.assertEqual(self._FAKE_SG_ID, rule_list[0]['security_group_id'])
+
+        # Test with remote_group_id
+        fake_sg_template['remote_group_id'] = self._FAKE_SG_ID
+        self._driver._sg_members[self._FAKE_SG_ID] = {self._FAKE_ETHERTYPE:
+                                                      [self._FAKE_MEMBER_IP]}
+        rule_list = self._driver._select_sg_rules_for_port(mock_port,
+                                                           'ingress')
+        self.assertEqual(self._FAKE_SG_ID, rule_list[0]['security_group_id'])
+        self.assertEqual('10.0.0.1/32', rule_list[0]['source_ip_prefix'])
+
+        # Test for fixed 'ip' existing in 'sg_members'
+        self._driver._sg_members[self._FAKE_SG_ID][self._FAKE_ETHERTYPE] = [
+            '10.0.0.2']
+
+        mock_port['fixed_ips'] = ['10.0.0.2']
+        rule_list = self._driver._select_sg_rules_for_port(mock_port,
+                                                           'ingress')
+        self.assertEqual([], rule_list)
+
+        # Test for 'egress' direction
+        fake_sg_template['direction'] = 'egress'
+        fix_ip = [self._FAKE_MEMBER_IP, '10.0.0.2']
+        self._driver._sg_members[self._FAKE_SG_ID][self._FAKE_ETHERTYPE] = (
+            fix_ip)
+
+        rule_list = self._driver._select_sg_rules_for_port(mock_port,
+                                                           'egress')
+        self.assertEqual('10.0.0.1/32', rule_list[0]['dest_ip_prefix'])
+
+        # Test for rules with a different direction
+        rule_list = self._driver._select_sg_rules_for_port(mock_port,
+                                                           'ingress')
+        self.assertEqual([], rule_list)
+
+    def test_update_security_group_rules(self):
+        mock_rule = [self._create_security_rule()]
+        self._driver.update_security_group_rules(self._FAKE_ID, mock_rule)
+        self.assertEqual(mock_rule,
+                         self._driver._sg_rule_templates[self._FAKE_ID])
+
+    def test_update_security_group_members(self):
+        mock_member = ['10.0.0.1/32']
+        self._driver.update_security_group_members(self._FAKE_ID, mock_member)
+        self.assertEqual(mock_member, self._driver._sg_members[self._FAKE_ID])
+
+    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
+                       '_select_sg_rules_for_port')
+    def test__generate_rules(self, mock_select_sg_rules):
+        mock_rule = [self._create_security_rule()]
+        mock_port = self._get_port()
+        mock_select_sg_rules.return_value = mock_rule
+        ports = self._driver._generate_rules([mock_port])
+
+        # Expected result
+        mock_rule.append(mock_rule[0])
+        expected = {self._FAKE_ID: mock_rule}
+        self.assertEqual(expected, ports)
+
+    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
+                       '_generate_rules')
+    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                        '_create_port_rules')
-    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriverMixin,
+    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                        '_add_sg_port_rules')
-    def test_prepare_port_filter(self, mock_add_rules, mock_create_rules):
+    def test_prepare_port_filter(self, mock_add_rules, mock_create_rules,
+                                 mock_gen_rules):
         mock_port = self._get_port()
         mock_create_default = self._driver._sg_gen.create_default_sg_rules
 
+        fake_rule = self._create_security_rule()
+        self._driver._get_rule_remote_address = mock.MagicMock(
+            return_value=self._FAKE_SOURCE_IP_PREFIX)
+        mock_gen_rules.return_value = {mock_port['id']: [fake_rule]}
+
         self._driver.prepare_port_filter(mock_port)
 
         self.assertEqual(mock_port,
                          self._driver._security_ports[self._FAKE_DEVICE])
-
+        mock_gen_rules.assert_called_with([self._driver._security_ports
+                                          [self._FAKE_DEVICE]])
         mock_add_rules.assert_called_once_with(
             self._FAKE_ID, mock_create_default.return_value)
-        self._driver._create_port_rules.assert_called_once_with(
-            self._FAKE_ID, mock_port['security_group_rules'])
 
-    def test_update_port_filter(self):
+        self._driver._create_port_rules.assert_called_with(
+            self._FAKE_ID, [fake_rule])
+
+    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
+                       '_generate_rules')
+    def test_update_port_filter(self, mock_gen_rules):
         mock_port = self._get_port()
         new_mock_port = self._get_port()
         new_mock_port['id'] += '2'
         new_mock_port['security_group_rules'][0]['ethertype'] += "2"
 
+        fake_rule_new = self._create_security_rule()
+        self._driver._get_rule_remote_address = mock.MagicMock(
+            return_value=self._FAKE_SOURCE_IP_PREFIX)
+
+        mock_gen_rules.return_value = {new_mock_port['id']: [fake_rule_new]}
+
         self._driver._security_ports[mock_port['device']] = mock_port
+        self._driver._sec_group_rules[new_mock_port['id']] = []
+
         self._driver._create_port_rules = mock.MagicMock()
         self._driver._remove_port_rules = mock.MagicMock()
         self._driver.update_port_filter(new_mock_port)
@@ -104,18 +200,24 @@ class TestHyperVSecurityGroupsDriverMixin(SecurityGroupRuleTestHelper):
         self._driver._remove_port_rules.assert_called_once_with(
             mock_port['id'], mock_port['security_group_rules'])
         self._driver._create_port_rules.assert_called_once_with(
-            new_mock_port['id'], new_mock_port['security_group_rules'])
+            new_mock_port['id'], [new_mock_port['security_group_rules'][0],
+                                  fake_rule_new])
         self.assertEqual(new_mock_port,
                          self._driver._security_ports[new_mock_port['device']])
 
-    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriverMixin,
+    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                        'prepare_port_filter')
     def test_update_port_filter_new_port(self, mock_method):
         mock_port = self._get_port()
-        self._driver.prepare_port_filter = mock.MagicMock()
-        self._driver.update_port_filter(mock_port)
+        new_mock_port = self._get_port()
+        new_mock_port['id'] += '2'
+        new_mock_port['device'] += '2'
+        new_mock_port['security_group_rules'][0]['ethertype'] += "2"
 
-        self._driver.prepare_port_filter.assert_called_once_with(mock_port)
+        self._driver._security_ports[mock_port['device']] = mock_port
+        self._driver.update_port_filter(new_mock_port)
+
+        self.assertNotIn(new_mock_port['device'], self._driver._security_ports)
 
     def test_remove_port_filter(self):
         mock_port = self._get_port()
@@ -125,10 +227,11 @@ class TestHyperVSecurityGroupsDriverMixin(SecurityGroupRuleTestHelper):
         self._driver.remove_port_filter(mock_port)
         self.assertNotIn(mock_port['device'], self._driver._security_ports)
         self.assertNotIn(mock_port['id'], self._driver._sec_group_rules)
+        self._driver._utils.clear_port_sg_acls_cache(mock_port['id'])
 
-    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriverMixin,
+    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                        '_add_sg_port_rules')
-    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriverMixin,
+    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                        '_remove_sg_port_rules')
     def test_create_port_rules(self, mock_remove, mock_add):
         mock_rule = mock.MagicMock()
@@ -145,7 +248,7 @@ class TestHyperVSecurityGroupsDriverMixin(SecurityGroupRuleTestHelper):
         mock_remove.assert_called_once_with(self._FAKE_ID, [mock_rule])
         mock_add.assert_called_once_with(self._FAKE_ID, [mock_rule])
 
-    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriverMixin,
+    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                        '_remove_sg_port_rules')
     def test_remove_port_rules(self, mock_remove):
         mock_rule = mock.MagicMock()
@@ -160,14 +263,27 @@ class TestHyperVSecurityGroupsDriverMixin(SecurityGroupRuleTestHelper):
     def test_add_sg_port_rules_exception(self):
         mock_rule = mock.MagicMock()
         self._driver._sec_group_rules[self._FAKE_ID] = []
-        self._driver._utils.create_security_rules.side_effect = Exception(
-            'Generated Exception for testing.')
+        self._driver._utils.create_security_rules.side_effect = (
+            exceptions.HyperVException(msg='Generated Exception for testing.'))
 
-        self._driver._add_sg_port_rules(self._FAKE_ID, [mock_rule])
+        self.assertRaises(exceptions.HyperVException,
+                          self._driver._add_sg_port_rules,
+                          self._FAKE_ID, [mock_rule])
 
         self.assertNotIn(mock_rule,
                          self._driver._sec_group_rules[self._FAKE_ID])
 
+    def test_add_sg_port_rules_port_not_found(self):
+        self._driver._sec_group_rules[self._FAKE_ID] = []
+        self._driver._utils.create_security_rules.side_effect = (
+            exceptions.NotFound(resource='port_id'))
+
+        self.assertRaises(exceptions.NotFound,
+                          self._driver._add_sg_port_rules,
+                          self._FAKE_ID, [mock.sentinel.rule])
+
+        self.assertNotIn(self._FAKE_ID, self._driver._sec_group_rules)
+
     def test_add_sg_port_rules(self):
         mock_rule = mock.MagicMock()
         self._driver._sec_group_rules[self._FAKE_ID] = []
@@ -184,19 +300,34 @@ class TestHyperVSecurityGroupsDriverMixin(SecurityGroupRuleTestHelper):
     def test_remove_sg_port_rules_exception(self):
         mock_rule = mock.MagicMock()
         self._driver._sec_group_rules[self._FAKE_ID] = [mock_rule]
-        self._driver._utils.remove_security_rules.side_effect = Exception(
-            'Generated Exception for testing.')
-        self._driver._remove_sg_port_rules(self._FAKE_ID, [mock_rule])
+        self._driver._utils.remove_security_rules.side_effect = (
+            exceptions.HyperVException(msg='Generated Exception for testing.'))
+
+        self.assertRaises(exceptions.HyperVException,
+                          self._driver._remove_sg_port_rules,
+                          self._FAKE_ID, [mock_rule])
 
         self.assertIn(mock_rule, self._driver._sec_group_rules[self._FAKE_ID])
 
+    def test_remove_sg_port_rules_port_not_found(self):
+        self._driver._sec_group_rules[self._FAKE_ID] = []
+        self._driver._utils.remove_security_rules.side_effect = (
+            exceptions.NotFound(resource='port_id'))
+
+        self.assertRaises(exceptions.NotFound,
+                          self._driver._remove_sg_port_rules,
+                          self._FAKE_ID, [mock.sentinel.rule])
+
+        self.assertNotIn(self._FAKE_ID, self._driver._sec_group_rules)
+
     def test_remove_sg_port_rules(self):
         mock_rule = mock.MagicMock()
         self._driver._sec_group_rules[self._FAKE_ID] = [mock_rule]
-        self._driver._remove_sg_port_rules(self._FAKE_ID, [mock_rule])
+        self._driver._remove_sg_port_rules(
+            self._FAKE_ID, [mock_rule, mock.sentinel.other_rule])
 
         self._driver._utils.remove_security_rules.assert_called_once_with(
-            self._FAKE_ID, [mock_rule])
+            self._FAKE_ID, [mock_rule, mock.sentinel.other_rule])
         self.assertNotIn(mock_rule,
                          self._driver._sec_group_rules[self._FAKE_ID])
 
@@ -296,7 +427,7 @@ class SecurityGroupRuleGeneratorR2TestCase(SecurityGroupRuleR2BaseTestCase):
         sg_rule1 = self._create_sg_rule(self._acl('protocol', 'tcp'))
         sg_rule2 = self._create_sg_rule(self._acl('protocol', 'udp'))
         sg_rule3 = self._create_sg_rule(self._acl('protocol', 'icmp'))
-        sg_rule4 = self._create_sg_rule(self._acl('protocol', 'icmpv6'))
+        sg_rule4 = self._create_sg_rule(self._acl('protocol', 'ipv6-icmp'))
 
         rule = self._create_security_rule()
         rule['protocol'] = sg_driver.ACL_PROP_MAP["default"]
@@ -376,8 +507,8 @@ class SecurityGroupRuleR2TestCase(SecurityGroupRuleR2BaseTestCase):
         sg_rule = self._create_sg_rule(self._acl('protocol', 'icmp'))
         self.assertFalse(sg_rule.Stateful)
 
-    def test_stateful_icmpv6(self):
-        sg_rule = self._create_sg_rule(self._acl('protocol', 'icmpv6'))
+    def test_stateful_ipv6_icmp(self):
+        sg_rule = self._create_sg_rule(self._acl('protocol', 'ipv6-icmp'))
         self.assertFalse(sg_rule.Stateful)
 
     def test_stateful_deny(self):
